import {
  ProtocolNotification,
  ProtocolRequest,
  ProtocolResponse,
} from './protocol';

import {
  PROTOCOL_NAME as RPCProtocol,
  RPCRequest,
  RPCResponse,
} from './protocol/rpc';

export type OnDisconnectFunc = () => void;
export type OnResponseFunc = (response: any) => void;

type ID_TYPE = number;

interface RequestQueue {
  resolve: (value?: any) => void;
  reject: (reason?: any) => void;
}

enum WebSocketStatus {
  NOT_CONNECT = 1,
  WILL_CONNECT,
  CONNECTED,
  WILL_DISCONNECT,
}

enum WebSocketReadyState {
  CONNECTING = 0,
  OPEN = 1,
  CLOSING = 2,
  CLOSED = 3,
}


export default class WebSocketRPC {
  private url: string;
  private connStatus: WebSocketStatus;
  private conn: WebSocket;
  private isReady: boolean;
  private requestID: ID_TYPE = 0;

  private onDisconnectListeners: OnDisconnectFunc[] = [];
  private onResponseListeners: Map<string, OnResponseFunc[]>;

  private requestQueue: Map<ID_TYPE, RequestQueue>;

  /**
   * name
   */
  public constructor(url: string) {
    this.isReady = false;
    this.connStatus = WebSocketStatus.NOT_CONNECT;
    this.url = url;
    this.onResponseListeners = new Map();
    this.requestQueue = new Map();
  }

  public Call(method: string, params: any[]): Promise<any> {


    return new Promise<any>((resolve, reject) => {
      const requestID = this.getRequestID();
      let request = new ProtocolRequest<ID_TYPE>(RPCProtocol, requestID);
      let req = new RPCRequest(method, params);
      request.setBody(req.toJSON());
      this.conn.send(request.toJSON());
      this.requestQueue.set(requestID, {resolve: resolve, reject: reject});
    });
  }

  public OnDisconnect(cb: OnDisconnectFunc): void {
    this.onDisconnectListeners.push(cb);
  }

  public initialize(): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        this.conn = new WebSocket(this.url);
      } catch(e) {
        console.log(e);
      }
      this.connStatus = WebSocketStatus.WILL_CONNECT;
      this.conn.onopen = (ev: Event): any => {
        this.connStatus = WebSocketStatus.CONNECTED;
        this.attachCallback();
        resolve();
        this.isReady = true;
        return null;
      };

      this.conn.onerror = (ev: Event): any => {
        this.connStatus = WebSocketStatus.NOT_CONNECT;
        reject(new Error('Cannot connect to API Server.'));
        return null;
      };
    });
  }

  private attachCallback(): void {
    this.conn.onopen = (ev: Event): any => {
      this.connStatus = WebSocketStatus.CONNECTED;
      this.isReady = true;
      return null;
    };

    this.conn.onclose = (ev: CloseEvent): any => {
      this.connStatus = WebSocketStatus.NOT_CONNECT;
      this.fireOnDisconnect();
      return null;
    };

    this.conn.onmessage = (ev: MessageEvent): any => {
      let response: ProtocolResponse<ID_TYPE> = JSON.parse(ev.data);
      let res: RPCResponse = JSON.parse(response.getBody());
      const requestID = response.getID();
      const error = response.getError();
      const result = res.getResult();

      if (this.requestQueue.has(requestID)) {
        let promise = this.requestQueue.get(requestID);
        this.requestQueue.delete(requestID);
        if (null != result) {
          promise.resolve(result);
        } else if (null != error) {
          promise.reject(error);
        } else {
          console.log('??????');
        }
      } else {
        // perhaps sever side send message
      }

      this.fireOnMessage(ev);
    };

    this.conn.onerror = (ev: Event): any => {
      console.log(ev);
      return null;
    };
  }


  private fireOnDisconnect(): void {
    for (let i = 0; i < this.onDisconnectListeners.length; i++) {
        this.onDisconnectListeners[i]();
    }
  }

  private fireOnMessage(e: MessageEvent): void {


    for (let i = 0; i < this.onDisconnectListeners.length; i++) {
        this.onDisconnectListeners[i]();
    }
  }

  private getRequestID(): number {
    return ++this.requestID;
  }

}
